@@ -32,18 +32,13 @@ __author_email__ = "biziqe@mathieu.fenniak.net"
 import logging
 import re
 import sys
+from collections.abc import Iterable, Sequence
 from io import BytesIO
 from math import ceil
 from typing import (
     Any,
     Callable,
-    Dict,
-    Iterable,
-    List,
     Optional,
-    Sequence,
-    Set,
-    Tuple,
     Union,
     cast,
 )
@@ -53,7 +48,6 @@ from .._utils import (
     WHITESPACES,
     StreamType,
     deprecation_no_replacement,
-    deprecation_with_replacement,
     logger_warning,
     read_non_whitespace,
     read_until_regex,
@@ -98,11 +92,10 @@ else:
     from typing_extensions import Self
 
 logger = logging.getLogger(__name__)
-NumberSigns = b"+-"
 IndirectPattern = re.compile(rb"[+-]?(\d+)\s+(\d+)\s+R[^a-zA-Z]")
 
 
-class ArrayObject(List[Any], PdfObject):
+class ArrayObject(list[Any], PdfObject):
     def replicate(
         self,
         pdf_dest: PdfWriterProtocol,
@@ -162,7 +155,7 @@ class ArrayObject(List[Any], PdfObject):
         """Emulate DictionaryObject.items for a list (index, object)."""
         return enumerate(self)
 
-    def _to_lst(self, lst: Any) -> List[Any]:
+    def _to_lst(self, lst: Any) -> list[Any]:
         # Convert to list, internal
         if isinstance(lst, (list, tuple, set)):
             pass
@@ -240,7 +233,7 @@ class ArrayObject(List[Any], PdfObject):
     def read_from_stream(
         stream: StreamType,
         pdf: Optional[PdfReaderProtocol],
-        forced_encoding: Union[None, str, List[str], Dict[int, str]] = None,
+        forced_encoding: Union[None, str, list[str], dict[int, str]] = None,
     ) -> "ArrayObject":
         arr = ArrayObject()
         tmp = stream.read(1)
@@ -263,12 +256,12 @@ class ArrayObject(List[Any], PdfObject):
             if peek_ahead == b"]":
                 break
             stream.seek(-1, 1)
-            # read and append obj
+            # read and append object
             arr.append(read_object(stream, pdf, forced_encoding))
         return arr
 
 
-class DictionaryObject(Dict[Any, Any], PdfObject):
+class DictionaryObject(dict[Any, Any], PdfObject):
     def replicate(
         self,
         pdf_dest: PdfWriterProtocol,
@@ -296,7 +289,7 @@ class DictionaryObject(Dict[Any, Any], PdfObject):
         except Exception:
             pass
 
-        visited: Set[Tuple[int, int]] = set()  # (idnum, generation)
+        visited: set[tuple[int, int]] = set()  # (idnum, generation)
         d__ = cast(
             "DictionaryObject",
             self._reference_clone(self.__class__(), pdf_dest, force_duplicate),
@@ -313,7 +306,7 @@ class DictionaryObject(Dict[Any, Any], PdfObject):
         pdf_dest: PdfWriterProtocol,
         force_duplicate: bool,
         ignore_fields: Optional[Sequence[Union[str, int]]],
-        visited: Set[Tuple[int, int]],  # (idnum, generation)
+        visited: set[tuple[int, int]],  # (idnum, generation)
     ) -> None:
         """
         Update the object from src.
@@ -336,8 +329,7 @@ class DictionaryObject(Dict[Any, Any], PdfObject):
                     del ignore_fields[x]
                     del ignore_fields[x]
                     continue
-                else:
-                    ignore_fields[x] -= 1  # type:ignore
+                ignore_fields[x] -= 1  # type:ignore
             x += 1
         #  First check if this is a chain list, we need to loop to prevent recur
         if any(
@@ -413,7 +405,7 @@ class DictionaryObject(Dict[Any, Any], PdfObject):
                         v.indirect_reference = None
                     vv = v.clone(pdf_dest, force_duplicate, ignore_fields)
                     assert vv.indirect_reference is not None
-                    self[k.clone(pdf_dest)] = vv.indirect_reference  # type: ignore[attr-defined]
+                    self[k.clone(pdf_dest)] = vv.indirect_reference
                 elif k not in self:
                     self[NameObject(k)] = (
                         v.clone(pdf_dest, force_duplicate, ignore_fields)
@@ -473,7 +465,7 @@ class DictionaryObject(Dict[Any, Any], PdfObject):
             raise ValueError("Key must be a PdfObject")
         if not isinstance(value, PdfObject):
             raise ValueError("Value must be a PdfObject")
-        return dict.setdefault(self, key, value)  # type: ignore
+        return dict.setdefault(self, key, value)
 
     def __getitem__(self, key: Any) -> PdfObject:
         return dict.__getitem__(self, key).get_object()
@@ -481,7 +473,7 @@ class DictionaryObject(Dict[Any, Any], PdfObject):
     @property
     def xmp_metadata(self) -> Optional[XmpInformationProtocol]:
         """
-        Retrieve XMP (Extensible Metadata Platform) data relevant to the this
+        Retrieve XMP (Extensible Metadata Platform) data relevant to this
         object, if available.
 
         See Table 347 — Additional entries in a metadata stream dictionary.
@@ -492,17 +484,14 @@ class DictionaryObject(Dict[Any, Any], PdfObject):
           return None if no metadata was found on the document root.
 
         """
-        from ..xmp import XmpInformation
+        from ..xmp import XmpInformation  # noqa: PLC0415
 
         metadata = self.get("/Metadata", None)
         if is_null_or_none(metadata):
             return None
+        assert metadata is not None, "mypy"
         metadata = metadata.get_object()
-
-        if not isinstance(metadata, XmpInformation):
-            metadata = XmpInformation(metadata)
-            self[NameObject("/Metadata")] = metadata
-        return metadata
+        return XmpInformation(metadata)
 
     def write_to_stream(
         self, stream: StreamType, encryption_key: Union[None, str, bytes] = None
@@ -512,7 +501,7 @@ class DictionaryObject(Dict[Any, Any], PdfObject):
                 "the encryption_key parameter of write_to_stream", "5.0.0"
             )
         stream.write(b"<<\n")
-        for key, value in list(self.items()):
+        for key, value in self.items():
             if len(key) > 2 and key[1] == "%" and key[-1] == "%":
                 continue
             key.write_to_stream(stream, encryption_key)
@@ -525,10 +514,10 @@ class DictionaryObject(Dict[Any, Any], PdfObject):
     def read_from_stream(
         stream: StreamType,
         pdf: Optional[PdfReaderProtocol],
-        forced_encoding: Union[None, str, List[str], Dict[int, str]] = None,
+        forced_encoding: Union[None, str, list[str], dict[int, str]] = None,
     ) -> "DictionaryObject":
         def get_next_obj_pos(
-            p: int, p1: int, rem_gens: List[int], pdf: PdfReaderProtocol
+            p: int, p1: int, rem_gens: list[int], pdf: PdfReaderProtocol
         ) -> int:
             out = p1
             for gen in rem_gens:
@@ -562,12 +551,12 @@ class DictionaryObject(Dict[Any, Any], PdfObject):
                 f"Dictionary read error at byte {hex(stream.tell())}: "
                 "stream must begin with '<<'"
             )
-        data: Dict[Any, Any] = {}
+        data: dict[Any, Any] = {}
         while True:
             tok = read_non_whitespace(stream)
             if tok == b"\x00":
                 continue
-            elif tok == b"%":
+            if tok == b"%":
                 stream.seek(-1, 1)
                 skip_over_comment(stream)
                 continue
@@ -619,7 +608,7 @@ class DictionaryObject(Dict[Any, Any], PdfObject):
         s = read_non_whitespace(stream)
         if s == b"s" and stream.read(5) == b"tream":
             eol = stream.read(1)
-            # odd PDF file output has spaces after 'stream' keyword but before EOL.
+            # Occasional PDF file output has spaces after 'stream' keyword but before EOL.
             # patch provided by Danial Sandler
             while eol == b" ":
                 eol = stream.read(1)
@@ -631,15 +620,14 @@ class DictionaryObject(Dict[Any, Any], PdfObject):
             if SA.LENGTH not in data:
                 if pdf is not None and pdf.strict:
                     raise PdfStreamError("Stream length not defined")
-                else:
-                    logger_warning(
-                        f"Stream length not defined @pos={stream.tell()}", __name__
-                    )
+                logger_warning(
+                    f"Stream length not defined @pos={stream.tell()}", __name__
+                )
                 data[NameObject(SA.LENGTH)] = NumberObject(-1)
             length = data[SA.LENGTH]
             if isinstance(length, IndirectObject):
                 t = stream.tell()
-                assert pdf is not None  # hint for mypy
+                assert pdf is not None, "mypy"
                 length = pdf.get_object(length)
                 stream.seek(t, 0)
             if length is None:  # if the PDF is damaged
@@ -654,7 +642,7 @@ class DictionaryObject(Dict[Any, Any], PdfObject):
             e = read_non_whitespace(stream)
             ndstream = stream.read(8)
             if (e + ndstream) != b"endstream":
-                # (sigh) - the odd PDF file has a length that is too long, so
+                # the odd PDF file has a length that is too long, so
                 # we need to read backwards to find the "endstream" ending.
                 # ReportLab (unknown version) generates files with this bug,
                 # and Python users into PDF files tend to be our audience.
@@ -680,10 +668,9 @@ class DictionaryObject(Dict[Any, Any], PdfObject):
             stream.seek(pos, 0)
         if "__streamdata__" in data:
             return StreamObject.initialize_from_dictionary(data)
-        else:
-            retval = DictionaryObject()
-            retval.update(data)
-            return retval
+        retval = DictionaryObject()
+        retval.update(data)
+        return retval
 
 
 class TreeObject(DictionaryObject):
@@ -770,8 +757,7 @@ class TreeObject(DictionaryObject):
             if "/Prev" in child_obj:
                 del child_obj["/Prev"]
             return child
-        else:
-            prev = cast("DictionaryObject", self["/Last"])
+        prev = cast("DictionaryObject", self["/Last"])
 
         while prev.indirect_reference != before:
             if "/Next" in prev:
@@ -846,14 +832,14 @@ class TreeObject(DictionaryObject):
 
         if NameObject("/Parent") not in child_obj:
             raise ValueError("Removed child does not appear to be a tree item")
-        elif child_obj[NameObject("/Parent")] != self:
+        if child_obj[NameObject("/Parent")] != self:
             raise ValueError("Removed child is not a member of this tree")
 
         found = False
         prev_ref = None
         prev = None
         cur_ref: Optional[Any] = self[NameObject("/First")]
-        cur: Optional[Dict[str, Any]] = cur_ref.get_object()  # type: ignore
+        cur: Optional[dict[str, Any]] = cur_ref.get_object()  # type: ignore
         last_ref = self[NameObject("/Last")]
         last = last_ref.get_object()
         while cur is not None:
@@ -881,8 +867,7 @@ class TreeObject(DictionaryObject):
         """Remove the object from the tree it is in."""
         if NameObject("/Parent") not in self:
             raise ValueError("Removed child does not appear to be a tree item")
-        else:
-            cast("TreeObject", self["/Parent"]).remove_child(self)
+        cast("TreeObject", self["/Parent"]).remove_child(self)
 
     def empty_tree(self) -> None:
         for child in self:
@@ -950,7 +935,7 @@ class StreamObject(DictionaryObject):
         pdf_dest: PdfWriterProtocol,
         force_duplicate: bool,
         ignore_fields: Optional[Sequence[Union[str, int]]],
-        visited: Set[Tuple[int, int]],
+        visited: set[tuple[int, int]],
     ) -> None:
         """
         Update the object from src.
@@ -984,7 +969,7 @@ class StreamObject(DictionaryObject):
             Hash considering type and value.
 
         """
-        # use of _data to prevent errors on non decoded stream such as JBIG2
+        # Use _data to prevent errors on non-decoded streams.
         return hash((super().hash_bin(), self._data))
 
     def get_data(self) -> bytes:
@@ -995,7 +980,7 @@ class StreamObject(DictionaryObject):
 
     def hash_value_data(self) -> bytes:
         data = super().hash_value_data()
-        data += self._data
+        data += self.get_data()
         return data
 
     def write_to_stream(
@@ -1012,15 +997,9 @@ class StreamObject(DictionaryObject):
         stream.write(self._data)
         stream.write(b"\nendstream")
 
-    @staticmethod
-    def initializeFromDictionary(data: Dict[str, Any]) -> None:
-        deprecation_with_replacement(
-            "initializeFromDictionary", "initialize_from_dictionary", "5.0.0"
-        )  # pragma: no cover
-
     @staticmethod
     def initialize_from_dictionary(
-        data: Dict[str, Any]
+        data: dict[str, Any]
     ) -> Union["EncodedStreamObject", "DecodedStreamObject"]:
         retval: Union[EncodedStreamObject, DecodedStreamObject]
         if SA.FILTER in data:
@@ -1035,7 +1014,7 @@ class StreamObject(DictionaryObject):
         return retval
 
     def flate_encode(self, level: int = -1) -> "EncodedStreamObject":
-        from ..filters import FlateDecode
+        from ..filters import FlateDecode  # noqa: PLC0415
 
         if SA.FILTER in self:
             f = self[SA.FILTER]
@@ -1079,7 +1058,7 @@ class StreamObject(DictionaryObject):
                 stops in your program.
 
         """
-        from ..filters import _xobj_to_image
+        from ..filters import _xobj_to_image  # noqa: PLC0415
 
         if self.get("/Subtype", "") != "/Image":
             try:
@@ -1101,27 +1080,26 @@ class EncodedStreamObject(StreamObject):
     def __init__(self) -> None:
         self.decoded_self: Optional[DecodedStreamObject] = None
 
-    # This overrides the parent method:
+    # This overrides the parent method
     def get_data(self) -> bytes:
-        from ..filters import decode_stream_data
+        from ..filters import decode_stream_data  # noqa: PLC0415
 
         if self.decoded_self is not None:
-            # cached version of decoded object
+            # Cached version of decoded object
             return self.decoded_self.get_data()
-        else:
-            # create decoded object
-            decoded = DecodedStreamObject()
 
-            decoded.set_data(decode_stream_data(self))
-            for key, value in list(self.items()):
-                if key not in (SA.LENGTH, SA.FILTER, SA.DECODE_PARMS):
-                    decoded[key] = value
-            self.decoded_self = decoded
-            return decoded.get_data()
+        # Create decoded object
+        decoded = DecodedStreamObject()
+        decoded.set_data(decode_stream_data(self))
+        for key, value in self.items():
+            if key not in (SA.LENGTH, SA.FILTER, SA.DECODE_PARMS):
+                decoded[key] = value
+        self.decoded_self = decoded
+        return decoded.get_data()
 
     # This overrides the parent method:
     def set_data(self, data: bytes) -> None:
-        from ..filters import FlateDecode
+        from ..filters import FlateDecode  # noqa: PLC0415
 
         if self.get(SA.FILTER, "") in (FT.FLATE_DECODE, [FT.FLATE_DECODE]):
             if not isinstance(data, bytes):
@@ -1162,17 +1140,13 @@ class ContentStream(DecodedStreamObject):
         self,
         stream: Any,
         pdf: Any,
-        forced_encoding: Union[None, str, List[str], Dict[int, str]] = None,
+        forced_encoding: Union[None, str, list[str], dict[int, str]] = None,
     ) -> None:
         self.pdf = pdf
-
-        # The inner list has two elements:
-        #  Element 0: List
-        #  Element 1: str
-        self._operations: List[Tuple[Any, bytes]] = []
+        self._operations: list[tuple[Any, bytes]] = []
 
         # stream may be a StreamObject or an ArrayObject containing
-        # multiple StreamObjects to be cat'd together.
+        # StreamObjects to be concatenated together.
         if stream is None:
             super().set_data(b"")
         else:
@@ -1255,7 +1229,7 @@ class ContentStream(DecodedStreamObject):
         except Exception:
             pass
 
-        visited: Set[Tuple[int, int]] = set()
+        visited: set[tuple[int, int]] = set()
         d__ = cast(
             "ContentStream",
             self._reference_clone(
@@ -1273,7 +1247,7 @@ class ContentStream(DecodedStreamObject):
         pdf_dest: PdfWriterProtocol,
         force_duplicate: bool,
         ignore_fields: Optional[Sequence[Union[str, int]]],
-        visited: Set[Tuple[int, int]],
+        visited: set[tuple[int, int]],
     ) -> None:
         """
         Update the object from src.
@@ -1296,7 +1270,7 @@ class ContentStream(DecodedStreamObject):
     def _parse_content_stream(self, stream: StreamType) -> None:
         # 7.8.2 Content Streams
         stream.seek(0, 0)
-        operands: List[Union[int, str, PdfObject]] = []
+        operands: list[Union[int, str, PdfObject]] = []
         while True:
             peek = read_non_whitespace(stream)
             if peek in (b"", 0):
@@ -1324,7 +1298,7 @@ class ContentStream(DecodedStreamObject):
             else:
                 operands.append(read_object(stream, None, self.forced_encoding))
 
-    def _read_inline_image(self, stream: StreamType) -> Dict[str, Any]:
+    def _read_inline_image(self, stream: StreamType) -> dict[str, Any]:
         # begin reading just after the "BI" - begin image
         # first read the dictionary of settings.
         settings = DictionaryObject()
@@ -1384,13 +1358,22 @@ class ContentStream(DecodedStreamObject):
 
         ei = stream.read(3)
         stream.seek(-1, 1)
-        if ei[0:2] != b"EI" or ei[2:3] not in WHITESPACES:
-            # Deal with wrong/missing `EI` tags.
+        if ei[:2] != b"EI" or ei[2:3] not in WHITESPACES:
+            # Deal with wrong/missing `EI` tags. Example: Wrong dimensions specified above.
             stream.seek(savpos, 0)
             data = extract_inline_default(stream)
+            ei = stream.read(3)
+            stream.seek(-1, 1)
+            if ei[:2] != b"EI" or ei[2:3] not in WHITESPACES:  # pragma: no cover
+                # Check the same condition again. This should never fail as
+                # edge cases are covered by `extract_inline_default` above,
+                # but check this ot make sure that we are behind the `EI` afterwards.
+                raise PdfStreamError(
+                    f"Could not extract inline image, even using fallback. Expected 'EI', got {ei!r}"
+                )
         return {"settings": settings, "data": data}
 
-    # This overrides the parent method:
+    # This overrides the parent method
     def get_data(self) -> bytes:
         if not self._data:
             new_data = BytesIO()
@@ -1412,20 +1395,20 @@ class ContentStream(DecodedStreamObject):
             self._data = new_data.getvalue()
         return self._data
 
-    # This overrides the parent method:
+    # This overrides the parent method
     def set_data(self, data: bytes) -> None:
         super().set_data(data)
         self._operations = []
 
     @property
-    def operations(self) -> List[Tuple[Any, bytes]]:
+    def operations(self) -> list[tuple[Any, bytes]]:
         if not self._operations and self._data:
             self._parse_content_stream(BytesIO(self._data))
             self._data = b""
         return self._operations
 
     @operations.setter
-    def operations(self, operations: List[Tuple[Any, bytes]]) -> None:
+    def operations(self, operations: list[tuple[Any, bytes]]) -> None:
         self._operations = operations
         self._data = b""
 
@@ -1436,7 +1419,7 @@ class ContentStream(DecodedStreamObject):
         elif self._data:
             self._data = b"q\n" + self._data + b"\nQ\n"
 
-    # This overrides the parent method:
+    # This overrides the parent method
     def write_to_stream(
         self, stream: StreamType, encryption_key: Union[None, str, bytes] = None
     ) -> None:
@@ -1448,54 +1431,51 @@ class ContentStream(DecodedStreamObject):
 def read_object(
     stream: StreamType,
     pdf: Optional[PdfReaderProtocol],
-    forced_encoding: Union[None, str, List[str], Dict[int, str]] = None,
+    forced_encoding: Union[None, str, list[str], dict[int, str]] = None,
 ) -> Union[PdfObject, int, str, ContentStream]:
     tok = stream.read(1)
     stream.seek(-1, 1)  # reset to start
     if tok == b"/":
         return NameObject.read_from_stream(stream, pdf)
-    elif tok == b"<":
+    if tok == b"<":
         # hexadecimal string OR dictionary
         peek = stream.read(2)
         stream.seek(-2, 1)  # reset to start
         if peek == b"<<":
             return DictionaryObject.read_from_stream(stream, pdf, forced_encoding)
-        else:
-            return read_hex_string_from_stream(stream, forced_encoding)
-    elif tok == b"[":
+        return read_hex_string_from_stream(stream, forced_encoding)
+    if tok == b"[":
         return ArrayObject.read_from_stream(stream, pdf, forced_encoding)
-    elif tok in (b"t", b"f"):
+    if tok in (b"t", b"f"):
         return BooleanObject.read_from_stream(stream)
-    elif tok == b"(":
+    if tok == b"(":
         return read_string_from_stream(stream, forced_encoding)
-    elif tok == b"e" and stream.read(6) == b"endobj":
+    if tok == b"e" and stream.read(6) == b"endobj":
         return NullObject()
-    elif tok == b"n":
+    if tok == b"n":
         return NullObject.read_from_stream(stream)
-    elif tok == b"%":
+    if tok == b"%":
         # comment
         skip_over_comment(stream)
         tok = read_non_whitespace(stream)
         stream.seek(-1, 1)
         return read_object(stream, pdf, forced_encoding)
-    elif tok in b"0123456789+-.":
+    if tok in b"0123456789+-.":
         # number object OR indirect reference
         peek = stream.read(20)
         stream.seek(-len(peek), 1)  # reset to start
         if IndirectPattern.match(peek) is not None:
-            assert pdf is not None  # hint for mypy
+            assert pdf is not None, "mypy"
             return IndirectObject.read_from_stream(stream, pdf)
-        else:
-            return NumberObject.read_from_stream(stream)
-    else:
-        pos = stream.tell()
-        stream.seek(-20, 1)
-        stream_extract = stream.read(80)
-        stream.seek(pos)
-        read_until_whitespace(stream)
-        raise PdfReadError(
-            f"Invalid Elementary Object starting with {tok!r} @{pos}: {stream_extract!r}"
-        )
+        return NumberObject.read_from_stream(stream)
+    pos = stream.tell()
+    stream.seek(-20, 1)
+    stream_extract = stream.read(80)
+    stream.seek(pos)
+    read_until_whitespace(stream)
+    raise PdfReadError(
+        f"Invalid Elementary Object starting with {tok!r} @{pos}: {stream_extract!r}"
+    )
 
 
 class Field(TreeObject):
@@ -1624,7 +1604,7 @@ class Destination(TreeObject):
         page: Union[NumberObject, IndirectObject, NullObject, DictionaryObject],
         fit: Fit,
     ) -> None:
-        self._filtered_children: List[Any] = []  # used in PdfWriter
+        self._filtered_children: list[Any] = []  # used in PdfWriter
 
         typ = fit.fit_type
         args = fit.fit_args
@@ -1657,12 +1637,12 @@ class Destination(TreeObject):
                 self[NameObject(TA.TOP)],
             ) = args
         elif typ in [TF.FIT_H, TF.FIT_BH]:
-            try:  # Preferred to be more robust not only to null parameters
+            try:  # Prefer to be more robust not only to null parameters
                 (self[NameObject(TA.TOP)],) = args
             except Exception:
                 (self[NameObject(TA.TOP)],) = (NullObject(),)
         elif typ in [TF.FIT_V, TF.FIT_BV]:
-            try:  # Preferred to be more robust not only to null parameters
+            try:  # Prefer to be more robust not only to null parameters
                 (self[NameObject(TA.LEFT)],) = args
             except Exception:
                 (self[NameObject(TA.LEFT)],) = (NullObject(),)
@@ -1711,8 +1691,8 @@ class Destination(TreeObject):
         return self.get("/Title")
 
     @property
-    def page(self) -> Optional[int]:
-        """Read-only property accessing the destination page number."""
+    def page(self) -> Optional[IndirectObject]:
+        """Read-only property accessing the IndirectObject of the destination page."""
         return self.get("/Page")
 
     @property
